#!/usr/bin/perl
# cbbs.cgi
########################################################################
#
# CBBS - Caste Bulletin Board System
# Version 1.0
#
# Script
#
# Copyright 2000 (c) N.Saito
#
########################################################################
#
# * UPDATE HISTORY *
#
# Date          Version Notes
# 2000/06/01     0.9b   First Beta.
# 2000/06/04     0.9.1  Fixed URI check.
# 2000/06/23     1.0    Release Version.
#
# * IMPORTANT NOTE *
#
# (Copyright and Escape Clause)
# 'Caste Bulletin Board System' is FREEWARE, but distribution is 
# not allowed without license.
# The creator of this program is not liable for any damage by using 
# this. 'Caste Bulletin Board System' is protected under the law and 
# the international treaty about copyright. All rights reserved.
#
#   Contact: N.Saito<ghb07031@nifty.ne.jp>
#            http://www.mycgiserver.com/~castebbs/
#
# (Must to do at first)
# At first, you should reconfigure this file. The path to the Perl 
# interpreter at your site may be different from mine. You must rewrite 
# Perl path in the first line of this script file. For example, if your 
# Perl path is '/usr/local/bin/perl', rewrite the first line of this 
# file like this:
#
#   #!/usr/local/bin/perl
#
# Next, you must setup the path to the configuration file. The 
# configuration file is a plane text file which holds many settings
# for your site. The configuration file is specified in this script
# as global variables, at just behind this so long comment section.
# "$configfile" is it. If you put your configuration file named 
# 'config.cfg' and put it the same directory as this file, you should 
# rewrite the variable as this:
#
#   $configfile = 'config.cfg';
#                           # ^ DON'T REMEBER end the line with 
#                           #   semi-colon(';').
#
# You may wanted to reconfigure the Administrator mode available. The
# administrator mode is the useful interface to setup miscellaneous
# elements, such User Informations, Programmable Authentication, 
# Trap Words, and so on. You have to rewrite $magicnumber as your own 
# password to enter the Administrator mode. The detail of this topic 
# will be told later, in the comment for the variables.
#
# (For Japanese User)
# If you want to operate Japanse DBCS code with this script, the 
# codeset of the template HTML file must be in "euc-jp" code because 
# some characters in the other codes, such as Shift-JIS, includes the 
# same value as ASCII code. This is said to other DBCS codeset. Use 
# appropreate characterset in your language. Maybe the codeset used in
# your language version UNIX will be safe.
#
# (Customizing Variables)
# To customize the script for your environment, simply change these
# variables:
#
# Name          Significant
# $magicnumber  String that indicates the *Magic Number* to enter
#               the Administrator mode. If you want to enter the Admin.
#               mode, simply type URI with the parametors like this:
#
#   http://your.domain/your_script_path.cgi?mode=your_magicnumber
#
#               But, be careful. Setting up the password for the 
#               Administrator mode may comes a great security hole.
#               If you are afraid to be cracked by this password,
#               leave this variable as blank.
# $configfile   String that contains the path to the configuration 
#               file. Relative path can be used and it is safer 
#               expression on the Web.
# $adminaddress String that indicates the mail-address of the 
#               administrator of the board which will be shown 
#               in the fatal error page. Don't remember escape the 
#               special character like at-mark('@') with back-slash, 
#               like this:
#
#   mail@admin.com => "mail\@admin.com"
#
# $anchoroptions String that will be inserted as options for the 
#               hyper-links of user's URLs and auto-anchors.

# Reconfigure these variables
####################################
$magicnumber = '';
$configfile = 'cbbs.cfg';
$adminaddress = '';
$anchoroptions = "target='_blank'";





########################################################################
# Main
########################################################################

# Initialyze variables
$accesstime = time();
$scriptname = 'Caste Bulletin Board System  version 1.0';
if ( rindex($ENV{'SCRIPT_NAME'},"/") < 0) {
 $scripturl = $ENV{'SCRIPT_NAME'};
} else {
 $scripturl = substr( $ENV{'SCRIPT_NAME'}, 
                       rindex($ENV{'SCRIPT_NAME'},"/")+1 );
}

# Configuration
%cfg = &LoadConfig( $configfile );
&CheckConfigError();
@castes = split( /,\s*/, $cfg{'CASTES'} );
@rkarma = split( /,\s*/, $cfg{'REQUISITE'} );
if ( @castes < $cfg{'NUMCASTE'} ) {
 &ErrorDie( "Misconfiguration found in $configfile: ".
            "Not enough castes", $cfg{'CONFIGERRSRC'}, "ERROR", );
}
if ( @rkarma < $cfg{'NUMCASTE'} ) {
 &ErrorDie( "Misconfiguration found in $configfile: ".
            "Not enough castes", $cfg{'CONFIGERRSRC'}, "ERROR", );
}

# Cookie
%cookie = &GetCookie();

# Parametors
%prm = &ParseCGIParams();

# Definition files
@users = &LoadUserProfile( $cfg{'USERS'} );
@atags = &LoadGeneralDefinition( $cfg{'TAGS'} ) if ( $cfg{'ALLOWTAGS'} );@trapwords = &LoadGeneralDefinition( $cfg{'TRAPWORDS'} );

# Enter appropreate mode
if ( ($prm{'mode'} eq $magicnumber) && $magicnumber ) { # Administrator
 &AdministratorMode();
} else {

 my( $view );
 if ( $prm{'mode'} eq 'alter' ) {
  $view = $cfg{'ALTERSRC'};
 } else {
  $view = $cfg{'MAINSRC'};
 }
 @articles = &LoadArticles( $cfg{'ARTICLELOG'} );
 &InitUser();

 if ( $prm{'mode'} eq 'reply' ) {

  # Reply Form mode
  if ( $ENV{'REQUEST_METHOD'} =~ /POST/i ) {
   &PostReply();
   &InitScope();
   &ShowArticles( $view );
  } else {
   &ReplyView();
  }

 } elsif ( ($prm{'mode'} eq 'control') ) {

  # Control Form mode
  if ( $ENV{'REQUEST_METHOD'} =~ /POST/i ) {
   &PostControl();
   &InitScope();
   &ShowArticles( $view );
  } else {
   &ControlView();
  }

 } else {

  # View mode
  &PostMain() if ( $ENV{'REQUEST_METHOD'} =~ /POST/i );
  &InitScope();
  &ShowArticles( $view );

 }

}
exit;




########################################################################
# Initialyze
########################################################################

# InitUser
####################################
sub InitUser {
 # Record access log
 &PushAccessLog( $cfg{'ACCESSLOG'} ) if ( $cfg{'ACCESSLOG'} );

 # Check Permission
 &ErrorForbidden() if ( !&AuthByEnvironment() );

 # Authentication
 if ( !$cfg{'AUTHTYPE'} ) {                     # Auth by Cookie
  if ( $cookie{'id'} ) {
   $userno = &FindUser( $cookie{'id'} );
  } else {
   $userno = -1;
  }
  if ( $userno < 0 ) {
   $$userinfo{'caste'} = $cfg{'STARTINGRANK'};
   $$userinfo{'karma'} = $rkarma[$caste[$cfg{'STARTINGRANK'}]];
  } else {
   $userinfo = $users[$userno];
  }
 } elsif  ( $cfg{'AUTHTYPE'} & 0x01 ) {         # Auth by Password
  if ( !$prm{'uid'} || !$prm{'pwd'} ) {
   &ConfigError("File error:$cfg{'AUTHFORM'}") if (!$cfg{'AUTHFORM'});
   my( %ins ) = ( '~ACCOUNT' => $cookie{'ACCOUNT'} );
   print &SetCookie( 'ACCOUNT', $$userinfo{'ua'} );
   print "Pragma: no-cache\n";
   print "Content-type: text/html\n\n";
   print &FootingPage(
      &InsertString( &LoadText($cfg{'AUTHFORM'}), %ins ) );
   exit;
  }
  $userno = &FindUser( $prm{'uid'} );
  if ( $userno < 0 ) {
   # User not found
   &ErrorForbidden() if ( $cfg{'AUTHTYPE'} & 0x04 );
   &GenerateNewID();
   $userno = $#users;
   $userinfo = $users[$userno];
   if ( $cfg{'AUTHTYPE'} & 0x02 ) {             # Auto register
    &InputError( $cfg{'ERRREGISTER'} ) 
                if ( ($prm{'uid'}=~/\W/) || ($prm{'pwd'}=~/\W/) );
    &InputError( $cfg{'ERRREGISTER'} ) if ( $prm{'uid'}<6 );
    &InputError( $cfg{'ERRREGISTER'} ) if ( $prm{'pwd'}<6 );
    $users[$#users]{'ua'} = $prm{'uid'};
    $users[$#users]{'up'} = $prm{'pwd'};
    &SaveUserProfile( $cfg{'USERS'}, *users );
   }
  } else {
   # User already exists
   $userinfo = $users[$userno];
   if ( $prm{'pwd'} ne $$userinfo{'up'} ) {
    &ErrorForbidden() if ( $cfg{'AUTHTYPE'} & 0x04 );
    &InputError( $cfg{'ERRDUPACCOUNT'} ) if ( $cfg{'AUTHTYPE'} & 0x02 );
   }
  }
  print &SetCookie( 'ACCOUNT', $$userinfo{'ua'} );
  $cookie{'ACCOUNT'} = $$userinfo{'ua'};
 }
}

# ParseCGIParams
####################################
sub ParseCGIParams {
 my( %result ) = ();
 my( @q ) = ();
 my( $k, $v );

 # for GET request
 @q = split( /&/, $ENV{'QUERY_STRING'} );
 foreach( @q ) {
  next if ( $_ !~ /([^=]+)=(.*)/ );
  $k=$1; $v = $2;
  $v =~ tr/+/ /;
  $v =~ s/%([0-9a-fA-F][0-9a-fA-F])/pack("C",hex($1))/eg;
  $result{$k} = $v;
 }

 # for POST request
 if ( $ENV{'REQUEST_METHOD'} =~ /POST/i ) {
  &InputError( $cfg{'ERRTOOLARGE'} )
          if ( $ENV{'CONTENT_LENGTH'}>$cfg{'MAXPOSTSIZE'} );
  my( $cmdline ) = "";
  sysread( STDIN, $cmdline, $ENV{'CONTENT_LENGTH'} );
  @q = split( /&/, $cmdline );
  foreach( @q ) {
   next if ( $_ !~ /([^=]+)=(.*)/ );
  $k=$1; $v = $2;
  $v =~ tr/+/ /;
  $v =~ s/%([0-9a-fA-F][0-9a-fA-F])/pack("C",hex($1))/eg;
  $result{$k} = $v;
  }
 }

 # Adjust Carriage Return
 $result{$k} =~ s/\r//g while ( ($k,$v)=each(%result) );

 return %result;
}

# AuthByEnvironment
####################################
sub AuthByEnvironment {
 return 1 if ( !$cfg{'LAWS'} );
 my( @d ) = &LoadGeneralDefinition( $cfg{'LAWS'} );
 my( $flag ) = 1;
 my( $target ) = '';
 foreach ( @d ) {
  if ( $_ =~ /^[Tt]:\s*(.*)/ ) {            # Target
   $target = $1;
  } elsif ( $_ =~ /^[Aa]:\s*(.*)/ ) {       # Allow
   $flag = 1 if ( $ENV{$target} =~ /$1/ );
  } elsif ( $_ =~ /^[Dd]:\s*(.*)/ ) {       # Deny
   $flag = 0 if ( $ENV{$target} =~ /$1/ );
  } else {
   &ConfigError( "Illegal definition found in $cfg{'LAWS'}" );
  }
 }
 return $flag;
}

# InitScope
####################################
sub InitScope {
 $page = $prm{'pg'};
 $minview = $cfg{'WIDTH'} * $page;
 $maxview = $minview + $cfg{'WIDTH'};
 $total = scalar( @articles );
 $maxview = $total if ( $maxview >= $total );
}




########################################################################
# Show Articles
########################################################################

# ShowArticles
####################################
sub ShowArticles {
 local( $text ) = &LoadText( $_[0] );
 my( %ins ) = &ExtractTemplate( *text );
 my( @tpl );
 $tpl[0] = $ins{'oppose'};
 $tpl[1] = $ins{'neutral'};
 $tpl[2] = $ins{'agree'};
 %ins = ();
 @articles = reverse( @articles );
 @articles = splice( @articles, $minview, $cfg{'WIDTH'} );
 my( $i );
 my( $result ) = "";
 $deapth = 1;
 for ( $i = 0 ; $i < @articles ; $i++ ) {
  %ins = &PrepareArticle( $articles[$i] );
  if ( $articles[$i]{'cn'} ) {
   $ins{'CHILD'} = &ShowChildren( $articles[$i]{'rep'}, @tpl );
  } else {
   $ins{'CHILD'} = '';
  }
  $result .= &InsertString( $tpl[$articles[$i]{'policy'}+1], %ins );
 }
 %ins = &PrepareElements();
 $ins{'ARTICLES'} = $result;

 print "Pragma: no-cache\n";
 print "Content-type: text/html\n\n";
 print &FootingPage( &InsertString( $text, %ins ) );
}

# ShowChildren
####################################
sub ShowChildren {
 my( $art ) = shift( @_ );
 my( @tpl  ) = @_;
 my( $i, %ins, $state );
 my( $result ) = "";
 $deapth ++;
 for ( $i = 0 ; $i < @$art ; $i++ ) {
  %ins = &PrepareArticle( $$art[$i] );
  if ( $$art[$i]{'cn'} ) {
   $ins{'CHILD'} = &ShowChildren( $$art[$i]{'rep'}, @tpl );
  } else {
   $ins{'CHILD'} = '';
  }
  $result .= &InsertString($tpl[$$art[$i]{'policy'}+1],%ins);
 }
 $deapth --;
 return $result;
}

# PrepareArticle
####################################
sub PrepareArticle {
 my( $art ) = @_;
 my( %ins ) = ();

 # Prepare normal values
 $ins{'CONTRIBUTOR'} = &EnsafeText( $$art{'name'} );
 if ( $$art{'st'} < 0 ) {       # Already Controled
  $ins{'SUBJECT'} = $cfg{'CONTROLEDSUBJECT'};
  $ins{'DESCRIPTION'} = '';
 } else {
  $ins{'SUBJECT'} = &EnsafeText( $$art{'sbj'} );
  $ins{'DESCRIPTION'} = &EnsafeText( $$art{'body'} );
 }
 if ( $$art{'caste'} < 0 ) {
  $ins{'CASTE'} = $cfg{'MASTERCASTE'};
 } else {
  $ins{'CASTE'} = $castes[$$art{'caste'}];
 }
 $ins{'POSTTIME'} = &TimeStamp( $$art{'time'}, 0x21 );
 $ins{'NUMCHILDREN'} = $$art{'cn'};

 # Prepare contributor with mailaddress
 if ( $$art{'mail'} ) {
  $ins{'CONTRIBUTOR'} = '<a href="mailto:'."$$art{'mail'}".'">'.
                        "$ins{'CONTRIBUTOR'}</a>";
 }

 # Prepare URL link
 if ( $$art{'url'} ) {
  $ins{'URL'} = '<a href="'.$$art{'url'}.'" $anchoroptions>';
  if ( $cfg{'URLTEXT'} ) {
   $ins{'URL'} .= $cfg{'URLTEXT'};
  } else {
   $ins{'URL'} .= &EnsafeText( $$art{'url'} );
  }
  $ins{'URL'} .= "</a>";
 } elsif( $cfg{'NOURLTEXT'} ) {
  $ins{'URL'} = $cfg{'NOURLTEXT'};
 }

 # Prepare REPLY link
 if ( ($deapth<$cfg{'MAXDEAPTH'}) || !$cfg{'MAXDEAPTH'} ) {
  $ins{'REPLY'} = '<a href="'."$scripturl?mode=reply".
                  "&a=$$art{'sn'}".'">'.
                  "$cfg{'REPLYTEXT'}</a>" if ( $$art{'st'}>=0 );
 } else {
  $ins{'REPLY'} = "";
 }

 # Prepare CONTROL link
 if ( $$art{'st'}>=0 ) {
  $ins{'CONTROL'} = '<a href="'.
                    "$scripturl?mode=control".
                    "&a=$$art{'sn'}".'">'.
                    "$cfg{'CONTROLTEXT'}</a>";
 } else {
  $ins{'CONTROL'} = "";
 }

 # Check URI in description
 if ( $cfg{'AUTOANCHOR'} ) {
  $ins{'DESCRIPTION'} =~ 
  s/([A-Za-z]\w*:)\/\/([\w~$-_@&!"',%=#\/\.\*\(\)\+\?]+)/<a href="$1\/\/$2" $anchoroptions>$1\/\/$2<\/a>/g;
 }

 # Check Tags in description
 foreach( @atags ) {
  $ins{'DESCRIPTION'} =~ s/&lt;(\/?$_)&gt;/<$1>/gi;
 }

 # Check Trapword in description
 foreach( @trapwords ) {
  $ins{'SUBJECT'} =~ 
             s/$_/$cfg{'TRAPMARKUP'}.$&.$cfg{'TRAPMARKUPOUT'}/eg;
  $ins{'DESCRIPTION'} =~ 
             s/$_/$cfg{'TRAPMARKUP'}.$&.$cfg{'TRAPMARKUPOUT'}/eg;
 }

 # Add system message to the description
 if ( $$art{'sysmsg'} ne '' && $cfg{'MSGMARKUPOUT'} ) {
  $ins{'DESCRIPTION'} .= "<br>\n$cfg{'MSGMARKUP'}".
                         "$$art{'sysmsg'}$cfg{'MSGMARKUPOUT'}";
 }

 return %ins;
}

# PrepareElements
####################################
sub PrepareElements {
 my( %ins ) = ();
 my( $maxpg ) = int( ( $total + $cfg{'WIDTH'} - 1 ) / $cfg{'WIDTH'} );

 $ins{'MINVIEW'} = $minview+1;
 $ins{'MAXVIEW'} = $maxview;
 $ins{'TOTAL'} = $total;

 # Prepare previous page link
 $ins{'PREVPAGE'} = '';
 if ( $page ) {
  my( $prev ) = $page - 1;
  $ins{'PREVPAGE'} = '<a href="'."$scripturl?pg=$prev".'">'.
                     "$cfg{'PREVPAGETEXT'}</a>";
 } elsif ( $cfg{'NOPREVTEXT'} ) {
  $ins{'PREVPAGE'} = $cfg{'NOPREVTEXT'};
 }

 # Prepare next page link
 $ins{'NEXTPAGE'} = '';
 if ( $maxpg-1 > $page ) {
  my( $next ) = $page + 1;
  $ins{'NEXTPAGE'} = '<a href="'."$scripturl?pg=$next".'">'.
                     "$cfg{'NEXTPAGETEXT'}</a>";
 } elsif ( $cfg{'NONEXTTEXT'} ) {
  $ins{'NEXTPAGE'} = $cfg{'NONEXTTEXT'};
 }

 # Prepare page links
 $ins{'PAGELINKS'} = "";
 for ( $i = 0 ; $i<$maxpg ; $i++ ) {
  if ( $i == $page ) {
   $ins{'PAGELINKS'} .= ($i+1).'&nbsp; ';
  } else {
   $ins{'PAGELINKS'} .= '<a href="'."$scripturl?pg=$i".'">'.
                        ($i+1).'</a>&nbsp; ';
  }
 }

 # Prepare Cookies
 foreach( ( 'CONTRIBUTOR', 'MAILADDRESS', 'URL', 'DELETEKEY',
            'ACCOUNT' ) ) {
  $ins{"~$_"} = $cookie{$_};
 }

 # Prepare Auth-By-Password
 $ins{'ACCOUNT'} = $$userinfo{'ua'};
 $ins{'PASSWORD'} = $$userinfo{'up'};

 # Prepare Caste
 if ( $$userinfo{'caste'}<0 ) {
  $ins{'CASTE'} = $cfg{'MASTERCASTE'};
 } else {
  $ins{'CASTE'} = $castes[$$userinfo{'caste'}];
 }

 return %ins;
}

# FootingPage
####################################
sub FootingPage {
 my( $text ) = @_;
 $debug =~ s/\n/<br>\n/g;
 my( $sign ) = "<br><hr>\n".
               "<a href='http://www.mycgiserver.com/~castebbs/'".
               " target='_blank'>\n".
               "$scriptname</a><br>\n".
               "Copyright 2000 (c) ".
               "<a href=\"mailto:ghb07031\@nifty.ne.jp\">N.Saito</a>".
               "<br>\n$debug";
 if ( $text !~ /<\/body>/i ) {
  if ( $text !~ /<\/html>/i ) {
   $text .= "$sign</body>\n</html>\n";
  } else {
   $text =~ s/(<\/html>)/$sign<\/body>$1/gi;
  }
 } else {
  $text =~ s/(<\/body>)/$sign$1/gi;
 }
 return $text;
}




########################################################################
# Article View
########################################################################

# PostMain
####################################
sub PostMain {
 return if ( &IsDuplicateArticle( \@articles ) );
 &CheckURI();
 &InputError( $cfg{'ERRNONAME'} ) if ( $prm{'contributor'} eq '' );
 &InputError( $cfg{'ERRNOSUBJECT'} ) if ( $prm{'subject'} eq '' );
 &InputError( $cfg{'ERRNOBODY'} ) if ( $prm{'description'} eq '' );
 if ( $prm{'deletekey'} ) {
  &InputError( $cfg{'ERRILLKEY'} ) if ( $prm{'deletekey'}=~/\W/ || 
                       length($prm{'deletekey'})<$cfg{'MINKEYLEN'} );
 }

 $deapth = 1;
 $prm{'policy'} = 0;
 my( $post ) = &CreateArticle( \@articles );
 $$userinfo{'karma'} += $cfg{'BASEKARMA'};
 $$userinfo{'name'} = $$post{'name'};
 $$userinfo{'mail'} = $$post{'mail'};
 &CheckDrop();
 &CheckTrap( $post );
 &CheckCaste( $$post{'name'}, $userinfo, $post );
 &ShiftArticles( $cfg{'STOCKLOG'}, *articles );
 &SaveArticles( $cfg{'ARTICLELOG'}, *articles );
 &SaveUserProfile( $cfg{'USERS'}, *users );
 &DetailAccessLog( $cfg{'ACCESSLOG'}, 'Article', $$post{'sn'} );
 my( $u ) = "$$post{'name'}($$userinfo{'id'})";
 $u .= "<$$post{'mail'}>" if ( $$post{'mail'} );
 &DetailAccessLog( $cfg{'ACCESSLOG'}, 'User', $u );
}





########################################################################
# Reply view
########################################################################

# ReplyView
####################################
sub ReplyView {
 $deapth = 0;
 my( $target ) = &FindArticle( $prm{'a'}, *articles );
 &InputError( $cfg{'ERRILLPARAM'} ) if ( !$target );

 local( $text ) = &LoadText( $cfg{'REPLYSRC'} );
 my( %tpl ) = &ExtractTemplate( *text );
 my( $i, %ins );
 my( $result ) = "";
 %ins = &PrepareArticle( $target );
 $result .= &InsertString( $tpl{'parent'}, %ins );
 %ins = &PrepareElements();
 $ins{'ARTICLES'} = $result;
 $ins{'SERIAL'} = $prm{'a'};
 $ins{'SUBJECT'} = $$target{'sbj'};
 foreach( @atags ) {
  $$target{'body'} =~ s/<(\/?$_)>//gi;
 }
 $ins{'SOURCEARTICLE'} = $cfg{'REPLYHEADER'}. $$target{'body'};
 $ins{'SOURCEARTICLE'} =~ s/\n/\n$cfg{'REPLYHEADER'}/g;

 print "Pragma: no-cache\n";
 print "Content-type: text/html\n\n";
 print &FootingPage( &InsertString( $text, %ins ) );
}

# PostReply
####################################
sub PostReply {
 return if ( &IsDuplicateArticle( \@articles ) );
 &CheckURI();
 $deapth = 0;
 my( $target ) = &FindArticle( $prm{'a'}, *articles );
 &InputError( $cfg{'ERRILLPARAM'} ) if ( !$target );
 &ErrorForbidden() if ( !$cfg{'ALLOWRESELF'} && 
        ($$target{'id'} eq $$userinfo{'id'}) );
 &InputError( $cfg{'ERRNONAME'} ) if ( $prm{'contributor'} eq '' );
 &InputError( $cfg{'ERRNOSUBJECT'} ) if ( $prm{'subject'} eq '' );
 &InputError( $cfg{'ERRNOBODY'} ) if ( $prm{'description'} eq '' );
 if ( $prm{'deletekey'} ) {
  &InputError( $cfg{'ERRILLKEY'} ) if ( $prm{'deletekey'}=~/\W/ || 
                       length($prm{'deletekey'})<$cfg{'MINKEYLEN'} );
 }

 $$target{'rep'} = [] if ( !$$target{'cn'} );
 my( $post ) = &CreateArticle( $$target{'rep'} );
 $$target{'cn'} = scalar( @{$$target{'rep'}} );
 $$target{'st'} = 1;
 $$userinfo{'karma'} += $cfg{'REPLYKARMA'};
 $$userinfo{'name'} = $$post{'name'};
 $$userinfo{'mail'} = $$post{'mail'};
 &CheckDrop();
 &CheckTrap( $post );
 &CheckCaste( $$post{'name'}, $userinfo, $post );
 my( $puno ) = &FindUser( $$target{'id'} );
 if ( $puno >= 0 ) {
  if ( $prm{'policy'}>0 ) {           # Agree
   $users[$puno]{'karma'} += $cfg{'AGREEKARMA'};
  } elsif ( $prm{'policy'}<0 ) {      # Oppose
   $users[$puno]{'karma'} += $cfg{'OPPOSEKARMA'};
  }
  &CheckCaste( $$target{'name'}, $users[$puno], $post );
 }
 &SaveArticles( $cfg{'ARTICLELOG'}, *articles );
 &SaveUserProfile( $cfg{'USERS'}, *users );
 &DetailAccessLog( $cfg{'ACCESSLOG'}, 'Article', $$post{'sn'} );
 &DetailAccessLog( $cfg{'ACCESSLOG'}, 'Reply To', $$target{'sn'} );
 my( $u ) = "$$post{'name'}($$userinfo{'id'})";
 $u .= "<$$post{'mail'}>" if ( $$post{'mail'} );
 &DetailAccessLog( $cfg{'ACCESSLOG'}, 'User', $u );
}




########################################################################
# Control View
########################################################################

# ControlView
####################################
sub ControlView {
 $deapth = 0;
 my( $target ) = &FindArticle( $prm{'a'}, *articles );
 &InputError( $cfg{'ERRILLPARAM'} ) if ( !$target );

 local( $text ) = &LoadText( $cfg{'CONTROLSRC'} );
 my( %tpl ) = &ExtractTemplate( *text );
 my( $i, %ins );
 my( $result ) = "";
 %ins = &PrepareArticle( $target );
 $result .= &InsertString( $tpl{'parent'}, %ins );
 %ins = &PrepareElements();
 $ins{'ARTICLES'} = $result;
 $ins{'SERIAL'} = $prm{'a'};

 print "Pragma: no-cache\n";
 print "Content-type: text/html\n\n";
 print &FootingPage( &InsertString( $text, %ins ) );
}

# PostControl
####################################
sub PostControl {
 &CheckURI();
 $deapth = 0;
 my( $target ) = &FindArticle( $prm{'a'}, *articles );
 &InputError( $cfg{'ERRILLPARAM'} ) if ( !$target );

 my( %ins );
 if ( $$userinfo{'caste'}<0 ) {             # By Admin.
  %ins = ( 'USER' => $cfg{'MASTERCASTE'} );
  $$target{'st'} = -2;
  $$target{'sysmsg'}.="<br>\n".&InsertString($cfg{'MSGDELOTHER'},%ins);
 } elsif ( $prm{'deletekey'} ) {                 # By Self
  &InputError( $cfg{'ERRDIFFKEY'} ) 
                          if ( $prm{'deletekey'} ne $$target{'key'} );
  %ins = ( 'USER' => $$target{'name'} );
  $$target{'st'} = -1;
  $$target{'sysmsg'}.="<br>\n".&InsertString($cfg{'MSGDELSELF'},%ins);
 } else {                                   # By Caste
  &InputError( $cfg{'ERRNORIGHT'} ) if ( !$cfg{'ISCONTROL'} );
  &InputError( $cfg{'ERRNONAME'} ) if ( $prm{'contributor'} eq '' );
  &InputError( $cfg{'ERRNORIGHT'} ) 
                     if ( $$target{'caste'}<=$$userinfo{'caste'} );
  %ins = ( 'USER' => $prm{'contributor'} );
  $$target{'st'} = -2;
  $$target{'sysmsg'}.="<br>\n".&InsertString($cfg{'MSGDELOTHER'},%ins);
  $$userinfo{'karma'} += $cfg{'DELETEKARMA'};
 }
 $$userinfo{'name'} = $prm{'contributor'};
 $$userinfo{'mail'} = $prm{'mailaddress'};
 &CheckDrop();
 &CheckCaste( $prm{'contributor'}, $userinfo, $target );
 &SaveArticles( $cfg{'ARTICLELOG'}, *articles );
 &SaveUserProfile( $cfg{'USERS'}, *users );
 &DetailAccessLog( $cfg{'ACCESSLOG'}, 'Delete', $$target{'sn'} );
 my( $u ) = "$prm{'contributor'}($$userinfo{'id'})";
 $u .= "<$prm{'mailaddress'}>" if ( $prm{'mailaddress'} );
 &DetailAccessLog( $cfg{'ACCESSLOG'}, 'User', $u );
}





########################################################################
# Post Article
########################################################################

# CreateArticle
####################################
sub CreateArticle {
 # Error Check
 my( @param ) = ( 'contributor', 'subject', 'description', 'policy' );
 my( @error ) = ( 'ERRNONAME', 'ERRNOSUBJECT', 'ERRNOBODY', 
                  'ERRNOPOLICY' );
 my( $i );
 for ( $i = 0 ; $i < @param ; $i++ ) {
  &InputError( $cfg{"$error[$i]"} ) if ( $prm{$param[$i]} eq "" );
 }

 # Register posted user
 if ( $userno<0 ) {
  &GenerateNewID();
  $userno = $#users;
  $userinfo = $users[$userno];
  print &SetCookie( 'id', $$userinfo{'id'} );
 }
 $$userinfo{'posttime'} = $accesstime;
 $$userinfo{'name'} = $prm{'contributor'};
 $$userinfo{'mail'} = $prm{'mailaddress'};

 # Create New Article
 my( $art ) = @_;
 if ( @$art ) {
  if ( substr( $$art[$#art]{'sn'}, 0, length($accesstime) ) 
  eq $accesstime ) {
   $$art[$#$art+1]{'sn'} = "$accesstime"."$deapth".
             substr("000".(substr($$art[$#art]{'sn'},-3)+1),-3);
  } else {
   $$art[$#$art+1]{'sn'} = "$accesstime"."$deapth"."000";
  }
 } else {
  $$art[$#$art+1]{'sn'} = "$accesstime"."$deapth"."000";
 }
 $$art[$#$art]{'cn'} = 0;
 $$art[$#$art]{'st'} = 0;
 $$art[$#$art]{'name'} = $prm{'contributor'};
 $$art[$#$art]{'mail'} = $prm{'mailaddress'};
 $$art[$#$art]{'sbj'} = $prm{'subject'};
 $$art[$#$art]{'body'} = $prm{'description'};
 $$art[$#$art]{'url'} = $prm{'url'};
 $$art[$#$art]{'time'} = $accesstime;
 $$art[$#$art]{'policy'} = $prm{'policy'};
 $$art[$#$art]{'sysmsg'} = "";
 $$art[$#$art]{'key'} = $prm{'deletekey'};
 $$art[$#$art]{'id'} = $$userinfo{'id'};
 $$art[$#$art]{'karma'} = $$userinfo{'karma'};
 $$art[$#$art]{'caste'} = $$userinfo{'caste'};

 print &SetCookie( 'CONTRIBUTOR', $prm{'contributor'} );
 print &SetCookie( 'MAILADDRESS', $prm{'mailaddress'} );
 print &SetCookie( 'URL', $prm{'url'} );
 print &SetCookie( 'DELETEKEY', $prm{'deletekey'} );
 $cookie{'CONTRIBUTOR'} = $prm{'contributor'};
 $cookie{'MAILADDRESS'} = $prm{'mailaddress'};
 $cookie{'URL'} = $prm{'url'};
 $cookie{'DELETEKEY'} = $prm{'deletekey'};

 return $$art[$#$art];
}

# IsDuplicateArticle
####################################
sub IsDuplicateArticle {
 my( $art ) = @_;
 foreach( @$art ) {
  &IsDuplicateArticle( $$_{'rep'} ) if ( $$_{'cn'} );
  next if ( $$_{'name'} ne $prm{'contributor'} );
  next if ( $$_{'sbj'} ne $prm{'subject'} );
  next if ( $$_{'body'} ne $prm{'description'} );
  return 1;
 }
 return 0;
}

# CheckURI
####################################
sub CheckURI {
 if ( $prm{'mailaddress'} ) {
  &InputError( $cfg{'ERRILLURI'} ) if (
    $prm{'mailaddress'} !~ /^[\w\/$-_@"#,%\.\*\(\)\+]+$/ );
 }
 if ( $prm{'url'} ) {
  &InputError( $cfg{'ERRILLURI'} ) if (
    $prm{'url'} !~ /^[A-Za-z]\w*:\/\/[\w~$-_@&!"',%=\/\.\*\(\)\+\?]+[#\w+]?$/ );
 }
}




########################################################################
# Load Files
########################################################################

# LoadGeneralDefinition
####################################
sub LoadGeneralDefinition {
 my( $file ) = @_;
 my( @result ) = ();
 open( IN, "<$file" ) || return @result;
# if ( $ENV{'WINDIR'} eq "" ) {          # In UNIX
  flock( IN, 1 );
# }
 while ( <IN> ) {
  chop();
  $_ =~ s/#.*$//g;              # remove comments
  $_ =~ s/^\s*//g;              # remove heading extra spaces
  $_ =~ s/\s*$//g;              # remove tailing extra spaces
  next if ( $_ eq "" );         # next if no descriptions in line.
  $_ =~ s/%([0-9a-fA-F][0-9a-fA-F])/pack("C",hex($1))/eg;
  push( @result, $_ );
 }
 close( IN );
 return @result;
}

# LoadText
####################################
sub LoadText {
 local( $file ) = @_;
 my( $result ) = "";
 open( IN, "<$file" ) || return $result;
 $result .= $_ while ( <IN> );
 close( IN );
 return $result;
}

# LoadConfig
####################################
sub LoadConfig {
 my( @f ) = &LoadGeneralDefinition( $_[0] );
 &ErrorDie( "Invalid Configuration File $_[0]" ) if ( !@f );
 my( %r ) = ();
 foreach ( @f ) {
  if ( $_ =~ /^(\w+)\s*=\s*/ ) {
   $r{$1} = $';
  } else {
   &ErrorDie( 
   "Misconfiguration found in $configfile: Syntax Error<br>$_" );
  }
 }
 if ( $r{'TRAPMARKUP'} ) {
  @f = split( /,\s*/, $r{'TRAPMARKUP'} );
  $r{'TRAPMARKUP'} = $f[0];
  $r{'TRAPMARKUPOUT'} = $f[1];
 }
 if ( $r{'MSGMARKUP'} ) {
  @f = split( /,\s*/, $r{'MSGMARKUP'} );
  $r{'MSGMARKUP'} = $f[0];
  $r{'MSGMARKUPOUT'} = $f[1];
 }
 return %r;
}




########################################################################
# Articles
########################################################################

# LoadArticles
####################################
sub LoadArticles {
 my( @result ) = ();
 my( @stack ) = ();
 my( $fn ) = "";
 my( $buf ) = "";
 my( $ref ) = \@result;
 push( @stack, $ref, 0 );
 open( IN, "<$_[0]" ) || return @result;
 while ( <IN> ) {
  if ( $_ =~ /^\f(\d+)\n$/ ) {
   $$ref[$#$ref+1]{'sn'} = $1;
   $fn = "";
  } elsif ( $_ =~ /\eEnd of Article/ ) {
   chop( $buf );
   $$ref[$#$ref]{$fn} = $buf;
   if ( $$ref[$#$ref]{'cn'} ) {
    $stack[$#stack]-- if ( $#stack>1 );
    push( @stack, $ref, $$ref[$#$ref]{'cn'} );
    $$ref[$#$ref]{'rep'} = [];
    $ref = $$ref[$#$ref]{'rep'};
   } elsif ( $stack[$#stack] ) {
    if ( !(--$stack[$#stack]) ) {
     pop( @stack );
     $ref = pop( @stack );
    }
   }
   $fn = "";
   next;
  } elsif ( $_ =~ /^\a(\w+)\n$/ ) {
   chop( $buf );
   $$ref[$#$ref]{$fn} = $buf if ( $fn );
   $buf = "";
   $fn = $1;
  } else {
   next if ( $fn eq "" );
   $buf .= $_;
  }
 }
 close( IN );
 return @result;
}

# SaveArticles
####################################
sub SaveArticles {
 my( $file, $art ) = @_;
# if ( $ENV{'WINDIR'} ne "" ) {          # In Windows
#  open( OUT, ">$file" ) || return 0;
# } else {                               # In UNIX
  open( OUT, "+<$file" ) || return 0;
  flock( OUT, 2 );
  truncate( OUT, 0 );
  seek( OUT, 0, 0 );
# }
 &WriteArticles( scalar(@$art), $art );
 close( OUT );
 return 1;
}

# WriteArticles
####################################
sub WriteArticles {
 my( $max, $art ) = @_;
 my( $i, $j );
 my( @p ) = ( 'st', 'name', 'mail', 'sbj', 'body', 'url', 'time',
              'policy', 'sysmsg', 'key', 'id', 'karma', 'caste' );
 for ( $i = 0 ; $i < $max ; $i++ ) {
  print OUT "\f$$art[$i]{'sn'}\n";
  print OUT "\acn\n";
  print OUT $$art[$i]{'cn'}."\n";
  for ( $j = 0 ; $j < @p ; $j++ ) {
   print OUT "\a$p[$j]\n";
   print OUT "$$art[$i]{$p[$j]}\n";
  }
  print OUT "\eEnd of Article\n";
  if ( @{$$art[$i]{'rep'}} ) {
   &WriteArticles( scalar(@{$$art[$i]{'rep'}}), $$art[$i]{'rep'} );
  }
 }
}

# ShiftArticles
####################################
sub ShiftArticles {
 return 1 if ( !$_[0] );
 my( $file, $art ) = @_;
 return 1 if ( scalar(@$art) <= $cfg{'WIDTH'}*$cfg{'MAXPAGE'} );
 open( OUT, "+<$file" ) || return 0;
# if ( !$ENV{'WINDIR'} ) {               # In UNIX
  flock( OUT, 2 );
# }
 seek( OUT, 0, 2 );
 &WriteArticles( 1, [shift( @$art )] );

 if ( (tell(OUT) > $cfg{'MAXSTOCKLOG'}) 
 && $cfg{'SENDMAIL'} && $cfg{'REPORTTO'} ) {
#  if ( !$ENV{'WINDIR'} ) {         # In UNIX
   open( PIPE, "| $cfg{'SENDMAIL'} $cfg{'REPORTTO'}" );
   print PIPE 
   "From: $scriptname\n".
   "To: $cfg{'REPORTTO'}\n".
   "Subject: Access Report ".&TimeStamp( $accesstime, 0x23 )."\n".
   "\n";
   seek( OUT, 0, 0 );
   print PIPE $_ while ( <OUT> );
   print PIPE "\n.\n";
   close( PIPE );
#  }
  seek( OUT, 0, 0 );
  truncate( OUT, 0 );
 }
 close( OUT );

 return 1;
}

# FindArticles
####################################
sub FindArticle {
 my( $serial, $art ) = @_;
 my( $i, $result );
 $deapth ++;
 for ( $i = 0 ; $i<@$art ; $i++ ) {
  return $$art[$i] if ( $$art[$i]{'sn'} eq $serial );
  next if ( !$$art[$i]{'cn'} );
  $result = &FindArticle( $serial, $$art[$i]{'rep'} );
  return $result if ( $result );
 }
 $deapth --;
 return 0;
}




########################################################################
# Text operation
########################################################################

# ExtractTemplate
####################################
sub ExtractTemplate {
 local( *src ) = @_;
 my( %result ) = ();
 my( $buf, $temp ) = "";
 while ( $src =~ /<template\b([^>]*)>/i ) {
  $src = $`;
  $buf = $1;
  $' =~ /<\/template>/i;
  $src .= $';
  $temp = $`;
  $result{$1} = $temp if ( $buf =~ /name=["]?([^"\s\n]+)["]?/ );
 }
 return %result;
}

# InsertString
####################################
sub InsertString {
 my( $src, %h ) = @_;
 my( $k, $v );
 $src =~ s/<!$k>/$v/gi while ( ($k,$v)=each(%h) );
 return $src;
}

# EnsafeText
####################################
sub EnsafeText {
 my( $text ) = @_;
 $text =~ s/&/&amp;/g;
 $text =~ s/"/&quot;/g;
 $text =~ s/</&lt;/g;
 $text =~ s/>/&gt;/g;
 $text =~ s/\n/<br>\n/g;
 return $text;
}

# TimeStamp
####################################
sub TimeStamp {
 my( $time, $fmt ) = @_;
 my( $type ) = ($fmt & 0xF0) >> 4;
 $fmt &= 0x0F;
 my( $ss, $mm, $hh, $day, $mon, $year, $dow, $a, $b );
 if ( $type == 2 ) {
  ( $ss, $mm, $hh, $day, $mon, $year, $dow, 
         $a, $b ) = gmtime( $time + $cfg{'TZMODIFER'}*3600 );
  $b = $cfg{'TZNAME'};
 } elsif ( $type == 1 ) {
  ($ss,$mm,$hh,$day,$mon,$year,$dow,$a,$b)=localtime( $time );
  $b = $ENV{'TZ'};
 } else {
  ($ss,$mm,$hh,$day,$mon,$year,$dow,$a,$b)=gmtime( $time );
  $b = "GMT";
 }
 $dow = ( "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" )[$dow];
 $year += 1900 if ( $year < 1900 );
 if ( $fmt == 2 ) {
  $mon ++;
  $mon = substr( "00$mon", -2 );
  $day = substr( "00$day", -2 );
  $hh = substr( "00$hh", -2 );
  $mm = substr( "00$mm", -2 );
  $ss = substr( "00$ss", -2 );
  $time = "$year/$mon/$day ($dow) $hh:$mm:$ss $b";
 } elsif ( $fmt == 1 ) {
  $mon ++;
  $time = substr($year,-2)."/$mon/$day($dow) $hh:$mm";
 } elsif ( $fmt == 3 ) {
  $mon ++;
  $time = substr($year,-2)."/$mon/$day($dow)";
 } else {
  $mon = ( "Jan", "Feb", "Mar", "Apr", "May", "Jun", 
           "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" )[$mon];
  $time = "$dow, $day-$mon-$year $hh:$mm:$ss GMT";
 }
 return $time;
}

# LineString
####################################
sub LineString {
 my( $p, $c ) = @_;
 my( $i );
 my( $r ) = "";
 for ( $i=0 ; $i<$c ; $i++ ) {
  $r .= $p;
 }
 return $r;
}

# GenericComments
####################################
sub GenericComments {
 my( $f, $t ) = @_;
 return "# $f\n".
 &LineString("#",72)."\n".
 "# \n# $scriptname\n# $t\n# \n".
 &LineString("#",72)."\n".
 "# \n# * WARNING *\n# \n".
 "# DO NOT change the contents of this file if you don't have \n".
 "# enough knowledge.\n# \n";
}




########################################################################
# Access Log
########################################################################

# PushAccessLog
####################################
sub PushAccessLog {
 my( $file, $detail ) = @_;
 open( OUT, "+<$file" ) || return 0;
# if ( !$ENV{'WINDIR'} ) {               # In UNIX
  flock( OUT, 2 );
# }
 seek( OUT, 0, 2 );
 if ( $detail ) {
  $detail .= "\n" if ( $detail !~ /\n$/ );
  print OUT $detail;
 } else {
  my( $k, $v );
  my( @ec ) = split( /,\s*/, $cfg{'ENVCONCERN'} );
  print OUT "\n<<$accesstime [".
            &TimeStamp( $accesstime, 0x22 ).
            "] $ENV{'REQUEST_METHOD'} $ENV{'REQUEST_URI'}\n";
  while ( ($k, $v)=each(%ENV) ) {
   foreach( @ec ) {
    next if ( $k !~ /$_/ );
    print OUT "$k=$v\n";
    last;
   }
  }
 }
 if ( (tell(OUT) > $cfg{'MAXACCESSLOG'}) 
 && $cfg{'SENDMAIL'} && $cfg{'REPORTTO'} ) {
#  if ( !$ENV{'WINDIR'} ) {         # In UNIX
   open( PIPE, "| $cfg{'SENDMAIL'} $cfg{'REPORTTO'}" );
   print PIPE 
   "From: $scriptname\n".
   "To: $cfg{'REPORTTO'}\n".
   "Subject: Access Report ".&TimeStamp( $accesstime, 0x23 )."\n".
   "\n";
   seek( OUT, 0, 0 );
   print PIPE $_ while ( <OUT> );
   print PIPE "\n.\n";
   close( PIPE );
#  }
  seek( OUT, 0, 0 );
  truncate( OUT, 0 );
 }
 close( OUT );
 return 1;
}

# DetailAccessLog
####################################
sub DetailAccessLog {
 my( $file, $n, $v ) = @_;
 return &PushAccessLog( $file, "$n: $v" );
}




########################################################################
# User Profile
########################################################################

# LoadUserProfile
####################################
sub LoadUserProfile {
 my( @result ) = ();
 my( @r ) = &LoadGeneralDefinition( $_[0] );
 return @result if ( !@r );
 my( @f, $i, $j );
 for ( $i = 0 ; $i < @r ; $i++ ) {
  @f = split( /,\s*/, $r[$i] );
  next if ( !$cfg{'AUTHTYPE'} && $i && 
            $accesstime>$f[2]+$cfg{'COOKIELIFE'}*86400 );   # timeout
  for ( $j = 0 ; $j < @f ; $j++ ) {
   $result[$i]{ ( 'serial', 'id', 'posttime', 'karma', 'caste', 'name', 
                  'mail', 'ua', 'up' )[$j] } = $f[$j];
  }
 }
 return @result;
}

# SaveUserProfile
####################################
sub SaveUserProfile {
 my( $file ) = shift( @_ );
 local( *users ) = shift( @_ );
# if ( $ENV{'WINDIR'} ne "" ) {          # In Windows
#  open( OUT, ">$file" ) || return 0;
# } else {                               # In UNIX
  open( OUT, "+<$file" ) || return 0;
  flock( OUT, 2 );
  truncate( OUT, 0 );
  seek( OUT, 0, 0 );
# }
 print OUT &GenericComments( $file, "User profile" )."# \n";
 print OUT "# serial,id,posttime,karma,caste,name,mail,acount,pwd\n\n";
 my( $i );
 for ( $i=0 ; $i<@users ; $i++ ) {
  print OUT $users[$i]{'serial'};
  foreach ( ( 'id', 'posttime', 'karma', 'caste', 'name', 'mail', 
              'ua', 'up' ) ) {
   $file = $users[$i]{$_};
   $file =~ s/(\W)/sprintf("%%%02X", unpack("C", $1))/eg;
   print OUT ",$file";
  }
  print OUT "\n";
 }
 close( OUT );
 return 1;
}

# GenerateNewID
####################################
sub GenerateNewID {
 return 0 if ( !@users );
 my( $id );
 $id = $users[$#users]{'id'};
 $users[$#users+1]{'serial'} = $users[$#users]{'serial'} + 1;
 if ( substr($users[$#users-1]{'id'},0,-3) eq $accesstime ) {
  $users[$#users]{'id'} = $accesstime.
                          substr( "000". 
                                (substr($users[$#users-1]{'id'},-3)+1),
                                  -3 );
 } else {
  $users[$#users]{'id'} = $accesstime."000";
 }
 $users[$#users]{'posttime'} = 0;
 $users[$#users]{'karma'} = 0;
 $users[$#users]{'caste'} = $cfg{'STARTINGRANK'};
 $users[$#users]{'name'} = "";
 $users[$#users]{'mail'} = "";
 $users[$#users]{'ua'} = "";
 $users[$#users]{'up'} = "";
 return $users[$#users];
}

# FindUser
####################################
sub FindUser {
 return -1 if ( !@users );
 my( $id ) = @_;
 my( $i );
 my( $f );
 if ( $cfg{'AUTHTYPE'} & 0x01 ) {       # Auth by pass
  $f = 'ua';
 } else {
  $f = 'id';
 }
 for ( $i = 0; $i < @users ; $i++ ) {
  return $i if ( $users[$i]{$f} eq $id );
 }
 return -1;
}




########################################################################
# Miscellaneous
########################################################################

# GetCaste
####################################
sub GetCaste {
 my( $i );
 for ( $i = 0 ; $i < @rkarma ; $i++ ) {
  return $i if ( $_[0]>=$rkarma[$i] );
 }
 return $#rkarma;
}

# CheckCaste
####################################
sub CheckCaste {
 my( $n, $u, $p ) = @_;
 return if ( $$u{'caste'} < 0 );         # Administrator
 my( $newc ) = &GetCaste( $$u{'karma'} );
 return if ( $$u{'caste'} == $newc );
 my( %ins ) = ( 'USER'      => &EnsafeText( $n ),
                'FROMCASTE' => $castes[$$u{'caste'}],
                'TOCASTE'   => $castes[$newc] );
 if ( $$u{'caste'} > $newc ) {
  $$p{'sysmsg'} .= "<br>\n".&InsertString( $cfg{'MSGGAIN'}, %ins );
 } elsif ( $$u{'caste'} < $newc ) {
  if ( $newc == $#castes ) {
   $$p{'sysmsg'} .= "<br>\n".&InsertString( $cfg{'MSGTRAP'}, %ins );
  } else {
   $$p{'sysmsg'} .= "<br>\n".&InsertString( $cfg{'MSGLOSE'}, %ins );
  }
 }
 $$u{'caste'} = $newc;
}

# CheckDrop
####################################
sub CheckDrop {
 return 0 if ( $$userinfo{'caste'} < 0 );         # Admin never drops
 return 0 if ( $$userinfo{'caste'} == $#castes );     # No more drop
 return 0 if ( $$userinfo{'posttime'}+$cfg{'DROPDELAY'}*86400 # 1day
             > $accesstime );   # Not drop
 if ( $$userinfo{'caste'} == $#castes-1 ) {         # Lowest rank
  $$userinfo{'karma'} = $rkarma[$#castes-1];
 } else {
  $$userinfo{'karma'} = $rkarma[$$userinfo{'caste'}+1];
 }
 return 1;
}

# CheckTrap
####################################
sub CheckTrap {
 return 0 if ( $$userinfo{'caste'} < 0 );         # Admin never trapped
 my( $p ) = @_;
 foreach ( @trapwords ) {
  if ( $$p{'sbj'}=~/$_/ || $$p{'body'}=~/$_/ ) {
   $$userinfo{'karma'} = $rkarma[$#rkarma];
   return 1;
  }
 }
 return 0;
}

# SetCookie
####################################
sub SetCookie {
 local( $k, $v ) = @_;
 &ErrorDie( "Misconfiguration found in $configfile: ".
            "Missing COOKIELIFE", $cfg{'CONFIGERRSRC'}, "ERROR", 
             ) if ( !defined($cfg{'COOKIELIFE'}) );
 my( $str ) = "Set-Cookie: ";
 $v =~ s/(\W)/sprintf("%%%02X",unpack("C",$1))/eg;
 $str .= "$k=$v; ";
 $str .= "expires=". &TimeStamp( $accesstime 
                      + $cfg{'COOKIELIFE'}*86400 );      # 86400sec=1day
 $str .= "; domain=". $ENV{'HTTP_HOST'};
 $str .= "; path=". $ENV{'SCRIPT_NAME'};
 $str .= ";\n";
 return $str;
}

# GetCookie
####################################
sub GetCookie {
 my( %c ) = ();
 return %c if ( !$ENV{'HTTP_COOKIE'} );
 my( $r, $k, $v );
 for $r ( split( /;\s*/, $ENV{'HTTP_COOKIE'} ) ) {
  ( $k, $v ) = split( /=/, $r );
  $v =~ s/%([0-9A-Fa-f][0-9A-Fa-f])/pack("C",hex($1))/eg;
  $c{$k} = $v;
 }
 return %c;
}




########################################################################
# Administrator Mode
########################################################################

# AdministratorMode
####################################
sub AdministratorMode {
 $adminurl .= "$scripturl?mode=$magicnumber";
 my( $result, $title, $i );

 # List of functions
 my( @commands ) = ( 
  'resetcookie',
  'viewaccesslog',
  'editusers',
  'edittrapwords',
  'edittags'
 );
 my( @functions ) = (
  \&AdminResetCookie,
  \&AdminViewAccessLog,
  \&AdminEditUsers,
  \&AdminEditTrapwords,
  \&AdminEditTags
 );
 my( @titles ) = (
  "Reset Administrator's ID",
  "View Access Log",
  "Edit Users",
  "Edit Trap Words",
  "Edit Tag Admitance"
 );
 my( @files ) = (
  'USERS',
  'ACCESSLOG',
  'USERS',
  'TRAPWORDS',
  'TAGS'
 );

 if ( $prm{'cmd'} ne "" ) {
  $title = "Error";
  $result = "Command error.<br>\n";
  for ( $i=0 ; $i<@commands ; $i++ ) {
   next if ( $prm{'cmd'} ne $commands[$i] );
   $admincmd = "$adminurl&cmd=$commands[$i]";
   $title = $titles[$i];
   $result = &{ $functions[$i] }();
   last;
  }
 } else {
  $title = "Menu";
  $result = 
  "This is administrator's menu to customize the behavior ".
  "of your BBS. Here is a list of the avaiable functions:<br>\n".
  "<ul type=circle>\n";
  for ( $i=0 ; $i<@commands ; $i++ ) {
   next if ( $files[$i] && !$cfg{$files[$i]} );
   $result .= '  <li><a href="'.
   "$adminurl&cmd=$commands[$i]".'">'.
   "$titles[$i]</a></li>\n";
  }
  $result .= "</ul>\n";
 }
 &AdminShowResult( $title, $result );
}

# AdminResetCookie
####################################
sub AdminResetCookie {
 my( $msg ) = "";
 if ( $prm{'status'} eq "execute" ) {       # Execute reset
  $msg =
  "Your administrator's identifier is successfully reseted.<br>\n".
  "The cookie will survive for 10 years.<br>\n";

  if ( @users ) {
   &GenerateNewID();
   $users[0]{'id'} = $users[$#users]{'id'};
   pop( @users );
  } else {
   $users[0]{'id'} = $accesstime."000";
  }
  $users[0]{'posttime'} = $accesstime;
  $users[0]{'serial'} = 0;
  $users[0]{'karma'} = 0;
  $users[0]{'caste'} = -1;
  $users[0]{'name'} = "";
  $users[0]{'mail'} = "";
  &SaveUserProfile( $cfg{'USERS'}, *users );

  print 
  "Set-Cookie: id=$users[0]{'id'}; expires=".
  &TimeStamp( $accesstime + 315360000 )."; ".
  "domain=". $ENV{'HTTP_HOST'}."; ".
  "path=".$ENV{'SCRIPT_NAME'}.";\n";

 } else {                                   # Introduction

  $msg = 
  "This function resets administrator's identifier and send new ".
  "cookie that keeps an appropreate ID on your local machine.<br>\n".
  'Click <a href="'."$admincmd&status=execute".
  '">HERE</a> to reset your '."
  administrator's identifier.<br>\n";

 }
 return $msg;
}

# AdminViewAccessLog
####################################
sub AdminViewAccessLog {
 my( %h, %h2, $r, $c );
 %h = ( 'MAGICNUMBER' => $magicnumber, 'FILE' =>$cfg{'ACCESSLOG'} );
 if ( $prm{'view'} ) {
  $h{'VIEWLINK'} = '<a href="'.$admincmd.'">'."\n";
  $h{'VIEWNAME'} = "resume simple";
 } else {
  $h{'VIEWLINK'} = '<a href="'."$admincmd&view=1".'">'."\n";
  $h{'VIEWNAME'} = "see details";
 }
 $h{'RECORDS'} = &tplAdminViewAccessLog1();
 open( IN, "<$cfg{'ACCESSLOG'}" );
 flock( IN, 1 ) if ( !$ENV{'WINDIR'} );
 $r = "";
 %h2 = ();
 $c = 0;
 while ( <IN> ) {
  chop();
  if ( $_ =~ /<<(\d+)\s+\[[^\]]+\]\s+(.*)/ ) {
   $h{'RECORDS'} .= &tplAdminViewAccessLog1() if ( !(++$c % 10) );
   $h{'RECORDS'}.= &InsertString( &tplAdminViewAccessLog2(),
                                  %h2 ) if ( %h2 );
   $h2{'COLOR'} = ('silver','white')[$c%2];
   $h2{'TIME'} = &TimeStamp( $1, 0x22 );
   $h2{'REQUEST'} = $2;
  } elsif ( ($_ ne "") && $prm{'view'} ) {
   $h2{'REQUEST'} .= "<br>\n".&EnsafeText($_);
  }
 }
 close( IN );
 $h{'RECORDS'}.= &InsertString( &tplAdminViewAccessLog2(), 
                                  %h2 ) if ( %h2 );
 return &InsertString( &tplAdminViewAccessLog0(), %h );
}

# AdminEditUsers
####################################
sub AdminEditUsers {
 my( %h, $i, %h2, %h3 );

 # Apply POST
 if ( ($ENV{'REQUEST_METHOD'} =~ /POST/i) && $prm{'entry'}<@users ) {
  if ( !$prm{'m'} ) {                # Modify User
   if ( $prm{'entry'} ) {
    if ( ($prm{'target'} eq 'k') && $prm{'k'}=~/^[\d-]+$/ ) {
     $users[$prm{'entry'}]{'karma'} = $prm{'k'};
     $users[$prm{'entry'}]{'caste'} = &GetCaste( $prm{'k'} );
    } elsif ( $prm{'target'} eq 'c' ) {
     $users[$prm{'entry'}]{'caste'} = $prm{'c'};
     $users[$prm{'entry'}]{'karma'} = $rkarma[ $prm{'c'} ];
    }
   }
   if ( $cfg{'AUTHTYPE'} & 0x01 ) {
    $users[$prm{'entry'}]{'ua'} = $prm{'a'};
    $users[$prm{'entry'}]{'up'} = $prm{'p'};
   }
  } elsif ( $prm{'m'} ==1 ) {       # Create User
   return "No account name.<br>" if ( $prm{'a'} eq '' );
   return "Invalid account.<br>" if ( $prm{'a'} =~ /(\W)/ );
   return "Too short account.<br>" if ( length($prm{'a'})<6 );
   return "Duplicate account.<br>" if ( &FindUser( $prm{'a'} )>=0 );
   return "No password.<br>" if ( $prm{'p'} eq '' );
   return "Invalid password.<br>" if ( $prm{'p'} =~ /(\W)/ );
   return "Too short password.<br>" if ( length($prm{'p'})<6 );
   &GenerateNewID();
   $users[$#users]{'ua'} = $prm{'a'};
   $users[$#users]{'up'} = $prm{'p'};
   if ( ($prm{'target'} eq 'k') && $prm{'k'}=~/^[\d-]+$/ ) {
    $users[$#users]{'karma'} = $prm{'k'};
    $users[$#users]{'caste'} = &GetCaste( $prm{'k'} );
   } elsif ( $prm{'target'} eq 'c' ) {
    $users[$#users]{'caste'} = $prm{'c'};
    $users[$#users]{'karma'} = $rkarma[ $prm{'c'} ];
   }
  } elsif ( $prm{'m'} ==2 ) {       # Delete User
   splice( @users, $prm{'entry'}, 1 ) if ( $prm{'entry'} );
  }
  &SaveUserProfile( $cfg{'USERS'}, *users );
 }

 # Prepare Global
 $h{'MAGICNUMBER'} = $magicnumber;
 $h{'FILE'} = $cfg{'USERS'};
 $h{'ACTION'} = $scripturl;
 $h{'CASTES'} = "";
 for ( $i = 0 ; $i < @castes ; $i++ ) {
  $h{'CASTES'} .= '<option value="'.$i.'">'.
                    "$castes[$i]($rkarma[$i])</option>\n";
 }
 $h{'METHODS'} = '<option value="0">Modify User</option>';
 if ( $cfg{'AUTHTYPE'} & 0x01 ) {       # Auth by password
  $h3{'NAMEFIELD'} = 'Account';
  $h3{'PASSFIELD'} = 'Password';
  $h{'METHODS'} .= '<option value="1">Create User</option>';
 } else {
  $h3{'NAMEFIELD'} = 'Name in Last Post';
  $h3{'PASSFIELD'} = 'Mail in Last Post';
 }
 $h{'METHODS'} .= '<option value="2">Delete User</option>';
 $h{'RECORDS'} = &InsertString( &tplAdminEditUsers1(), %h3 );

 # Insert Admin
 %h2 = %{$users[0]};
 $h2{'caste'} = $cfg{'MASTERCASTE'};
 $h2{'color'} = 'white';
 $h2{'posttime'} = &TimeStamp( $h2{'posttime'}, 0x21 );
 if ( $cfg{'AUTHTYPE'} &0x01 ) {
  $h2{'name'} = $h2{'ua'};
  $h2{'mail'} = $h2{'up'};
 } else {
  $h2{'name'} = &EnsafeText( $h2{'name'} );
 }
 $h{'RECORDS'} .= &InsertString( &tplAdminEditUsers2(), %h2 );

 # Insert User
 for ( $i = 1 ; $i < @users ; $i++ ) {
  if ( !($i % 10) ) {
   $h{'RECORDS'} .= &InsertString( &tplAdminEditUsers1(), %h3 );
  }
  %h2 = %{$users[$i]};
  $h2{'caste'} = $castes[$h2{'caste'}];
  $h2{'color'} = ('white','silver')[$i%2];
  $h2{'posttime'} = &TimeStamp( $h2{'posttime'}, 0x21 );
  if ( $cfg{'AUTHTYPE'} &0x01 ) {
   $h2{'name'} = $h2{'ua'};
   $h2{'mail'} = $h2{'up'};
  } else {
   $h2{'name'} = &EnsafeText( $h2{'name'} );
  }
  $h{'RECORDS'} .= &InsertString( &tplAdminEditUsers2(), %h2 );
 }
 return &InsertString( &tplAdminEditUsers0(), %h );
}

# AdminEditTrapwords
####################################
sub AdminEditTrapwords {
 if ( !$cfg{'TRAPWORDS'} ) {
  return "The Trapwords file name is not specified.".
         "Please setup file name first.<br>\n";
 }
 if ( $ENV{'REQUEST_METHOD'} =~ /POST/i ) {
#  if ( $ENV{'WINDIR'} ne "" ) {          # In Windows
#   open( OUT, ">$cfg{'TRAPWORDS'}" ) || return 0;
#  } else {                               # In UNIX
   open( OUT, "+<$cfg{'TRAPWORDS'}" ) || return 0;
   flock( OUT, 2 );
   truncate( OUT, 0 );
   seek( OUT, 0, 0 );
#  }
  my( $t ) = $prm{'trapword'};
  print OUT &GenericComments( $cfg{'TRAPWORDS'}, 
            "Trap Words" )."# \n";
  print OUT "#\n# DON'T REMEMBER escape out these characters:\n";
  print OUT "#   + * . ? | \ ( ) [ ] { }\n#\n\n";
  $t =~ s/#/%23/g;
  print OUT $t;
  close( OUT );
 }
 my( @trapwords ) = &LoadGeneralDefinition( $cfg{'TRAPWORDS'} );
 my( %h, $i );
 %h = ( 'MAGICNUMBER' => $magicnumber, 'FILE' => $cfg{'TRAPWORDS'},
        'ACTION' => $scripturl, 'TRAPWODS' => '' );
 for ( $i = 0 ; $i < @trapwords ; $i++ ) {
  $h{'TRAPWORDS'} .= "$trapwords[$i]\n";
 }
 return &InsertString( &tplAdminEditTrapwords(), %h );
}

# AdminEditTags
####################################
sub AdminEditTags {
 if ( !$cfg{'TAGS'} ) {
  return "The tags file name is not specified.".
         "Please setup file name first.<br>\n";
 }
 if ( !$cfg{'ALLOWTAGS'} ) {
  return "The present configuration does not allow to use tags.".
         "Please setup the configuration to allow first.<br>\n";
 }
 my( @tags ) = &LoadGeneralDefinition( $cfg{'TAGS'} );
 my( $i, $t );
 if ( $ENV{'REQUEST_METHOD'} =~ /POST/i ) {
  if ( $prm{'op'}==1 ) {
   push( @tags, $prm{'tag'} ) if ( $prm{'tag'} ne "" );
  } elsif ( $prm{'op'}==2 ) {
   splice( @tags, $prm{'target'}-1, 1 )
           if ( 0<$prm{'target'} && $prm{'target'}<scalar(@tags)+1 );
  }
#  if ( $ENV{'WINDIR'} ne "" ) {          # In Windows
#   open( OUT, ">$cfg{'TAGS'}" ) || return 0;
#  } else {                               # In UNIX
   open( OUT, "+<$cfg{'TAGS'}" ) || return 0;
   flock( OUT, 2 );
   truncate( OUT, 0 );
   seek( OUT, 0, 0 );
#  }
  print OUT &GenericComments( $cfg{'TAGS'}, "Tag Admitance" )."# \n";
  for ( $i = 0 ; $i < @tags ; $i++ ) {
   $t = $tags[$i];
   $t =~ s/#/%23/g;
   print OUT "$t\n";
  }
  close( OUT );
 }
 my( %h, %h2 );
 %h = ( 'MAGICNUMBER' => $magicnumber, 'FILE' => $cfg{'TAGS'},
        'ACTION' => $scripturl, 'RECORDS' => '' );
 for ( $i=0 ; $i<@tags ; $i++ ) {
  $h{'RECORDS'} .= &tplAdminEditTags1() if ( !($i%10) );
  $h2{'NUM'} = $i+1;
  $h2{'TAG'} = $tags[$i];
  $h2{'color'} = ('white','silver')[$i%2];
  $h{'RECORDS'} .= &InsertString( &tplAdminEditTags2(), %h2 );
 }
 return &InsertString( &tplAdminEditTags0(), %h );
}

# AdminShowResult
####################################
sub AdminShowResult {
 my( %h ) = ( 'TITLE'=>$_[0], 'RESULT'=>$_[1], 
              'CHARSET'=>$cfg{'DBCSCHARSET'}, 'SCRIPTNAME'=>$scriptname,
              'BBSURL'=>"<a href=\"$scripturl\">", 
              'ADMINURL'=>"<a href=\"$adminurl\">" );
 print "Content-Type: text/html\n\n";
 print &InsertString( &tplAdminGeneral(), %h );
}




########################################################################
# Errorcheck
########################################################################

# ErrorDie
####################################
sub ErrorDie {
 local( $msg, $file, $rep ) = @_;
 print "Content-type: text/html\n\n";
 my( $text ) = "";
 $text = &LoadText( $file ) if ( ($file ne "") and ($rep ne "") );

 if ( $text ne "" ) {               # error template is exist
  $rep = "<!".$rep.">";
  $text =~ s/$rep/$msg/gi;
  print $text;
 } else {
  my( %h ) = ( 'MSG'=>$msg, 'SCRIPTNAME'=>$scriptname );
  if ( $adminaddress ) {
   $h{'MAILTOADMIN'} = '<a href="'.$adminaddress.'">'.
                       "administrator</a>";
  } else {
   $h{'MAILTOADMIN'} = "administrator";
  }
  print &InsertString( &tplFatalError(), %h );
 }
 exit;
}

# ErrorForbidden
####################################
sub ErrorForbidden {
 print "Content-type: text/html\nStatus: 403 Forbidden\n\n";
 print &tplForbidden();
 exit;
}

# ShowError
####################################
sub ShowError {
 my( $text ) = &LoadText( $_[1] );
 my( %ins ) = ( 'ERROR'=>$_[0] );
 print "Content-type: text/html\n\n";
 print &FootingPage( &InsertString( $text, %ins ) );
 exit;
}

# InputError
####################################
sub InputError {
 &ShowError( $_[0], $cfg{'INPUTERRSRC'} );
}

# ConfigError
####################################
sub ConfigError {
 &ShowError( $_[0], $cfg{'CONFIGERRSRC'} );
}

# CheckConfigError
####################################
sub CheckConfigError {
 my( @err ) = 
 (
  'MAINSRC', 
  'INPUTERRSRC', 
  'CONFIGERRSRC', 
  'ARTICLELOG', 
  'USERS', 
  'WIDTH', 
  'MAXPAGE', 
  'ISCONTROL', 
  'AUTOANCHOR', 
  'ALLOWTAGS', 
  'NUMCASTES', 
  'CASTES', 
  'REQUISITE', 
  'MASTERCASTE', 
  'STARTINGRANK', 
  'BASEKARMA', 
  'REPLYKARMA', 
  'AGREEKARMA', 
  'OPPOSEKARMA', 
  'DELETEKARMA', 
  'PREVPAGETEXT', 
  'NEXTPAGETEXT', 
  'ERRNONAME', 
  'ERRNOSUBJECT', 
  'ERRNOBODY', 
  'ERRNOPOLICY', 
  'ERRILLURI', 
  'ERRDIFFKEY', 
  'ERRNORIGHT', 
  'ERRILLPARAM', 
  'ERRILLKEY', 
  'ERRTOOLARGE'
 );
 foreach ( @err ) {
  &ErrorDie( "Misconfiguration found in $configfile:".
             " $_ is not specified." ) if ( $cfg{$_} eq '' );
 }
 $cfg{'DROPDELAY'} = $cfg{'COOKIELIFE'} if ( !$cfg{'DROPDELAY'} );
 $cfg{'REPLYHEADER'} = '#' if ( !$cfg{'REPLYHEADER'} );
}




########################################################################
# Templates
########################################################################

# Fatal Error
####################################
sub tplFatalError {
 return <<TPL_FATAL_ERROR;
<html>
<head>
<meta http-equiv="Pragma" content="no-cache">
<title>Fatal Error</title>
</head>
<body>
<h1>Fatal Error<hr></h1>
The program failed to complete your request because unexpected
error occured.Plese contact your <!MAILTOADMIN>.<br>
<br>
<table border=1 width=100% cellspacing=1 cellpadding=2>
<caption><strong>Reason:</strong></caption>
 <tr><td><!MSG></td></tr>
</table>
<hr>
<!SCRIPTNAME><br>
</body>
</html>
TPL_FATAL_ERROR
}

# Forbidden
####################################
sub tplForbidden {
 return <<TPL_FORBIDDEN;
<html>
<head>
<meta http-equiv="Pragma" content="no-cache">
<title>403 Forbidden</title>
</head>
<body>
<h1>Forbidden<hr></h1>
You don't have permission to access.<br>
</body>
</html>
TPL_FORBIDDEN
}

# Administrator mode main
####################################
sub tplAdminGeneral {
 return <<TPL_ADMIN_GENERAL;
<html>
<head>
<meta http-equiv="Content-type" content="text/html; charset=<!CHARSET>">
<meta http-equiv="Pragma" content="no-cache">
<title>Administrator Mode/<!TITLE> [<!SCRIPTNAME>]</title>
</head>
<body bgcolor=white text=black>
<table width=100% border=4 cellspacing=2 cellpadding=8><tr><td>
 <big><strong><!SCRIPTNAME><br>Administrator Mode</strong></big>
</td></tr></table>
<table width=100% border=0 cellspacing=0 cellpadding=48><tr><td>
 <table width=100% border=0 cellspacing=0 cellpadding=1 bgcolor=black>
  <tr><td>
   <table width=100% border=0 cellspacing=0 cellpadding=4 bgcolor=white>
    <tr><td><big><strong><!TITLE></strong></big></td></tr>
   </table>
  </td></tr>
 </table>
 <table width=100% border=0 cellspacing=0 cellpadding=24><tr><td>

<!-------------------------------------------------------------------->
<!RESULT>
<!-------------------------------------------------------------------->

 </td></tr></table>
</td></tr></table>
<hr>
<table width=100% border=0 cellspacing=0 cellpadding=0><tr>
 <td width=100%><font color=white>.</font></td>
 <td nowrap><!BBSURL>Back to BBS</a></td>
 <td nowrap><font color=white>____</font></td>
 <td nowrap><!ADMINURL>Back to administrator menu</td>
</tr></table>
</body>
</html>
TPL_ADMIN_GENERAL
}


# Administrator mode - Access Log
####################################
sub tplAdminViewAccessLog0 {
 return <<TPL_ADMIN_VIEW_ACCESSLOG0;
This function shows the access log to your BBS.<br>
Click <!VIEWLINK>HERE</a> to <!VIEWNAME>.<br>
<br>
<strong>File</strong> = <!FILE><br>
<br>
<table width=100% border=0 cellspacing=1 cellpadding=2>
<!RECORDS>
 <tr bgcolor=black><td colspan=2></td></tr>
</table>
</form>
TPL_ADMIN_VIEW_ACCESSLOG0
}

sub tplAdminViewAccessLog1 {
 return <<TPL_ADMIN_VIEW_ACCESSLOG1;
 <tr bgcolor=black>
  <td nowrap><font color=white>Acess Time</font></td>
  <td nowrap width=100%><font color=white>Request</font></td>
 </tr>
TPL_ADMIN_VIEW_ACCESSLOG1
}

sub tplAdminViewAccessLog2 {
 return <<TPL_ADMIN_VIEW_ACCESSLOG2;
 <tr bgcolor=<!COLOR>>
  <td nowrap><!TIME></td>
  <td>
<!REQUEST>
  </td>
 </tr>
TPL_ADMIN_VIEW_ACCESSLOG2
}


# Administrator mode - Edit Users
####################################
sub tplAdminEditUsers0 {
 return <<TPL_ADMIN_EDIT_USERS0;
This function provides an interface to edit user's status.<br>
To edit a user, select 'Modify User', check a type between Karma 
and Caste, enter appropreate value by the corresponding control, 
and choose a user to edit listed in below.Finally, click [Submit] 
to apply changes.<br>
To delete a user, select 'Delete User', choose a user to delete, and 
click [Submit].<br>
&lt;Only in Authentication by Password&gt;To create a new user, select 
'Create User', check a type a type between Karma and Caste, enter 
appropreate value by the corresponding control, enter new user's 
account name and password in the editbox, and then click [Submit].<br>
'Account Name' and 'Password' box works only in Authentication by 
Password mode. They're applied for 'Modify User' command, too.<br>
<br>
<br>
<strong>File</strong> = <!FILE><br>
<br>
<form action="<!ACTION>" method="POST">
<input type=hidden name="mode" value="<!MAGICNUMBER>">
<input type=hidden name="cmd" value="editusers">
<table width=100% border=4 cellspacing=2 cellpadding=2><tr><td>
 <table width=100% border=0 cellspacing=0 cellpadding=4>
  <tr>
   <td nowrap>
    <input type=radio name="target" value="k">
    Karma
   </td>
   <td><input type=edit name="k" value="" maxlength=8></td>
   <td nowrap>
    <input type=radio name="target" value="c">
    Caste
   </td>
   <td>
    <select name="c">
     <!CASTES>
    </select>
   </td>
   <td width=100%><font color=white>.</font></td>
  </tr>
  <tr>
   <td nowrap>Account Name:</td>
   <td><input type=edit name="a" value="" maxlength=20></td>
   <td nowrap>Password:</td>
   <td><input type=edit name="p" value="" maxlength=20></td>
   <td width=100%><font color=white>.</font></td>
  </tr>
  <tr>
   <td>
    <select name="m">
     <!METHODS>
    </select>
   </td>
   <td width=100%><font color=white>.</font></td>
   <td nowrap><input type=submit>&nbsp;<input type=reset></td>
  </tr>
 </table>
</td></tr></table>
<br>
<table width=100% border=0 cellspacing=1 cellpadding=2>
<!RECORDS>
 <tr bgcolor=black><td colspan=7></td></tr>
</table>
</form>
TPL_ADMIN_EDIT_USERS0
}

sub tplAdminEditUsers1 {
 return <<TPL_ADMIN_EDIT_USERS1;
 <tr bgcolor=black>
  <td align=right nowrap><font color=white>Serial</font></td>
  <td nowrap><font color=white>ID</font></td>
  <td><font color=white>Latest Post</font></td>
  <td align=right nowrap><font color=white align=right>Karma</font></td>
  <td nowrap><font color=white>Caste</font></td>
  <td><font color=white><!NAMEFIELD></font></td>
  <td><font color=white><!PASSFIELD></font></td>
 </tr>
TPL_ADMIN_EDIT_USERS1
}

sub tplAdminEditUsers2 {
 return <<TPL_ADMIN_EDIT_USERS2;
 <tr bgcolor=<!COLOR>>
  <td align=right nowrap>
   <input type=radio name="entry" value="<!SERIAL>">
   <!SERIAL>
  </td>
  <td nowrap><!ID></td>
  <td><!POSTTIME></td>
  <td align=right nowrap><!KARMA></td>
  <td><!CASTE></td>
  <td><!NAME></td>
  <td><!MAIL></td>
 </tr>
TPL_ADMIN_EDIT_USERS2
}


# Administrator mode - EditTrapwords
####################################
sub tplAdminEditTrapwords {
 return <<TPL_ADMIN_EDIT_TRAPWORDS;
This function provides an interface to edit Trap Words.<br>
The Trap Words are the definitions of words that traps someone 
who post the article with any of them. If an user enter the 
description including a Trap Word, he immediately drops his
Caste into the worst one:)<br>
The definition of the Trap Words is a simple text that contains 
the word to trap in each line. And the Trap Word is described by the
rule called 'Regular Expression' which is a popular pattern matching
method.<br>
If you don't know mach about Regular Expression, simply note that 
the very some characters, such plus('<strong>+</strong>'), asterisk
('<strong>*</strong>'), question('<strong>?</strong>'), vertical bar
('<strong>|</strong>'), back-slash('<strong>\\</strong>'), parentheses
('<strong>(</strong>' and '<strong>)</strong>'), brackets
('<strong>[</strong>' and '<strong>]</strong>') and curly braces
('<strong>{</strong>' and '<strong>}</strong>'), are regarded as
meta-character, and they must be escaped out with back-slash
('<strong>\\</strong>'). For example, if you want to register a word 
"<u>(in paren!)</u>", you need to write "<u>\\(in paren!\\)</u>". Also 
back-slash itself will be escaped out as "\\\\".<br>
<br>
<strong>File</strong> = <!FILE><br>
<br>
<form action="<!ACTION>" method="POST">
<input type=hidden name="mode" value="<!MAGICNUMBER>">
<input type=hidden name="cmd" value="edittrapwords">
<table width=100% border=4 cellspacing=2 cellpadding=2><tr><td>
 <table width=100% border=0 cellspacing=0 cellpadding=4>
  <tr><td align=center>
   <textarea name="trapword" cols=80% rows=10><!TRAPWORDS></textarea>
  </td></tr>
  <tr><td nowrap align=right>
   <input type=submit>&nbsp;<input type=reset>
  </td></tr>
 </table>
</td></tr></table>
<br>
</form>
TPL_ADMIN_EDIT_TRAPWORDS
}


# Administrator mode - Edit Tag Admission
####################################
sub tplAdminEditTags0 {
 return <<TPL_ADMIN_EDIT_TAGS0;
This function provides an interface to edit the Tag Admitance.<br>
The Tag Admitance is a list that determine which HTML tag is allowed to 
be used in the article.<br>
To append a new tag to allow, enter the tag to allow(with out &lt; or &gt;),
select 'Append', and then click [Submit] button.<br>
To delete an existing tag to prevent to be used, select 'Delete', check the 
number to delete, and click [Submit].<br>
<br>
<strong>File</strong> = <!FILE><br>
<br>
<form action="<!ACTION>" method="POST">
<input type=hidden name="mode" value="<!MAGICNUMBER>">
<input type=hidden name="cmd" value="edittags">
<table width=100% border=4 cellspacing=2 cellpadding=2><tr><td>
 <table width=100% border=0 cellspacing=0 cellpadding=4>
  <tr>
   <td nowrap>Append Tag:</td>
   <td width=100%>
    <input type=text name="tag" value="" size=50% maxlength=256>
   </td>
   <td align=right rowspan=2 nowrap>
    <input type=submit>&nbsp;<input type=reset>
   </td>
  </tr>
  <tr>
   <td nowrap>Operation:</td>
   <td><select name="op">
    <option value="1">Append</option>
    <option value="2">Delete</option>
   </select></td>
  </td>
 </table>
</td></tr></table>
<br>
<table width=100% border=0 cellspacing=1 cellpadding=2>
<!RECORDS>
 <tr bgcolor=black><td colspan=3></td></tr>
</table>
</form>
TPL_ADMIN_EDIT_TAGS0
}


sub tplAdminEditTags1 {
 return <<TPL_ADMIN_EDIT_TAGS1;
 <tr bgcolor=black>
  <td align=right nowrap><font color=white>No.</font></td>
  <td align=center nowrap><font color=white>Tag</font></td>
  <td><font color=white>Sample</font></td>
 </tr>
TPL_ADMIN_EDIT_TAGS1
}


sub tplAdminEditTags2 {
 return <<TPL_ADMIN_EDIT_TAGS2;
 <tr bgcolor=<!COLOR>>
  <td align=right nowrap>
   <input type=radio name="target" value="<!NUM>"><!NUM>
  </td>
  <td align=center nowrap>&lt;<!TAG>&gt;&nbsp;-&nbsp;&lt;/<!TAG>&gt;</td>
  <td nowrap>
   <table cellspacing=0 cellpadding=0 border=0><tr>
    <td><<!TAG>>Sample</td>
    <td width=100% nowrap>&nbsp;of the effect</<!TAG>></td>
   </tr></table>
  </td>
  
 </tr>
TPL_ADMIN_EDIT_TAGS2
}





